# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import json
from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.quicksight import QuickSightCreateIngestionOperator
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.quicksight import QuickSightSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

"""
Prerequisites:
1. The account which runs this test must manually be activated in Quicksight here:
https://quicksight.aws.amazon.com/sn/console/signup?#
2. The activation process creates an IAM Role called `aws-quicksight-service-role-v0`.
 You have to add a policy named 'AWSQuickSightS3Policy' with the S3 access permissions.
 The policy name is enforced, and the permissions json can be copied from `AmazonS3FullAccess`.

NOTES:  If Create Ingestion fails for any reason, that ingestion name will remain in use and
 future runs will stall with the sensor returning a status of QUEUED "forever".  If you run
 into this behavior, changing the template for the ingestion name or the ENV_ID and re-running
 the test should resolve the issue.
"""

DAG_ID = "example_quicksight"

sys_test_context_task = SystemTestContextBuilder().build()

SAMPLE_DATA_COLUMNS = ["Project", "Year"]
SAMPLE_DATA = """'Airflow','2015'
    'OpenOffice','2012'
    'Subversion','2000'
    'NiFi','2006'
"""


@task
def get_aws_account_id() -> int:
    return boto3.client("sts").get_caller_identity()["Account"]


@task
def create_quicksight_data_source(
    aws_account_id: str, datasource_name: str, bucket: str, manifest_key: str
) -> str:
    response = boto3.client("quicksight").create_data_source(
        AwsAccountId=aws_account_id,
        DataSourceId=datasource_name,
        Name=datasource_name,
        Type="S3",
        DataSourceParameters={
            "S3Parameters": {"ManifestFileLocation": {"Bucket": bucket, "Key": manifest_key}}
        },
    )
    return response["Arn"]


@task
def create_quicksight_dataset(aws_account_id: int, dataset_name: str, data_source_arn: str) -> None:
    table_map = {
        "default": {
            "S3Source": {
                "DataSourceArn": data_source_arn,
                "InputColumns": [{"Name": name, "Type": "STRING"} for name in SAMPLE_DATA_COLUMNS],
            }
        }
    }

    boto3.client("quicksight").create_data_set(
        AwsAccountId=aws_account_id,
        DataSetId=dataset_name,
        Name=dataset_name,
        PhysicalTableMap=table_map,
        ImportMode="SPICE",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_quicksight_data_source(aws_account_id: str, datasource_name: str):
    boto3.client("quicksight").delete_data_source(AwsAccountId=aws_account_id, DataSourceId=datasource_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_dataset(aws_account_id: str, dataset_name: str):
    boto3.client("quicksight").delete_data_set(AwsAccountId=aws_account_id, DataSetId=dataset_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_ingestion(aws_account_id: str, dataset_name: str, ingestion_name: str) -> None:
    client = boto3.client("quicksight")
    with contextlib.suppress(client.exceptions.ResourceNotFoundException):
        # suppress ResourceNotFoundException: Ingestion has already terminated on its own.
        client.cancel_ingestion(
            AwsAccountId=aws_account_id,
            DataSetId=dataset_name,
            IngestionId=ingestion_name,
        )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    account_id = get_aws_account_id()

    env_id = test_context[ENV_ID_KEY]
    bucket_name = f"{env_id}-quicksight-bucket"
    data_filename = "sample_data.csv"
    dataset_id = f"{env_id}-data-set"
    datasource_id = f"{env_id}-data-source"
    ingestion_id = f"{env_id}-ingestion"
    manifest_filename = f"{env_id}-manifest.json"
    manifest_contents = {"fileLocations": [{"URIs": [f"s3://{bucket_name}/{data_filename}"]}]}

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=bucket_name)

    upload_manifest_file = S3CreateObjectOperator(
        task_id="upload_manifest_file",
        s3_bucket=bucket_name,
        s3_key=manifest_filename,
        data=json.dumps(manifest_contents),
        replace=True,
    )

    upload_sample_data = S3CreateObjectOperator(
        task_id="upload_sample_data",
        s3_bucket=bucket_name,
        s3_key=data_filename,
        data=SAMPLE_DATA,
        replace=True,
    )

    data_source = create_quicksight_data_source(
        aws_account_id=account_id,
        datasource_name=datasource_id,
        bucket=bucket_name,
        manifest_key=manifest_filename,
    )

    create_dataset = create_quicksight_dataset(account_id, dataset_id, data_source)

    # [START howto_operator_quicksight_create_ingestion]
    create_ingestion = QuickSightCreateIngestionOperator(
        task_id="create_ingestion",
        data_set_id=dataset_id,
        ingestion_id=ingestion_id,
    )
    # [END howto_operator_quicksight_create_ingestion]

    # QuickSightCreateIngestionOperator waits by default, setting as False to test the Sensor below.
    create_ingestion.wait_for_completion = False

    # If this sensor appears to freeze with a "QUEUED" status, see note above.
    # [START howto_sensor_quicksight]
    await_job = QuickSightSensor(
        task_id="await_job",
        data_set_id=dataset_id,
        ingestion_id=ingestion_id,
    )
    # [END howto_sensor_quicksight]
    await_job.poke_interval = 10

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        account_id,
        create_s3_bucket,
        upload_manifest_file,
        upload_sample_data,
        data_source,
        create_dataset,
        # TEST BODY
        create_ingestion,
        await_job,
        # TEST TEARDOWN
        delete_dataset(account_id, dataset_id),
        delete_quicksight_data_source(account_id, datasource_id),
        delete_ingestion(account_id, dataset_id, ingestion_id),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.links.logs import CloudWatchEventsLink
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS

from unit.amazon.aws.links.test_base_aws import BaseAwsLinksTestCase

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult


class TestCloudWatchEventsLink(BaseAwsLinksTestCase):
    link_class = CloudWatchEventsLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "us-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "awslogs_region": "ap-southeast-2",
                    "awslogs_group": "/test/logs/group",
                    "awslogs_stream_name": "test/stream/d56a66bb98a14c4593defa1548686edf",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/cloudwatch/home"
                "?region=ap-southeast-2#logsV2:log-groups/log-group/%2Ftest%2Flogs%2Fgroup"
                "/log-events/test%2Fstream%2Fd56a66bb98a14c4593defa1548686edf"
            ),
            region_name="us-west-1",
            aws_partition="aws",
            awslogs_region="ap-southeast-2",
            awslogs_group="/test/logs/group",
            awslogs_stream_name="test/stream/d56a66bb98a14c4593defa1548686edf",
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.hooks.bedrock import (
    BedrockAgentHook,
    BedrockAgentRuntimeHook,
    BedrockHook,
    BedrockRuntimeHook,
)


class TestBedrockHooks:
    @pytest.mark.parametrize(
        "test_hook, service_name",
        [
            pytest.param(BedrockHook(), "bedrock", id="bedrock"),
            pytest.param(BedrockRuntimeHook(), "bedrock-runtime", id="bedrock-runtime"),
            pytest.param(BedrockAgentHook(), "bedrock-agent", id="bedrock-agent"),
            pytest.param(BedrockAgentRuntimeHook(), "bedrock-agent-runtime", id="bedrock-agent-runtime"),
        ],
    )
    def test_bedrock_hooks(self, test_hook, service_name):
        assert test_hook.conn is not None
        assert test_hook.conn.meta.service_model.service_name == service_name



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.models import BaseOperator
from airflow.providers.amazon.aws.hooks.s3 import S3Hook

if TYPE_CHECKING:
    from airflow.utils.context import Context


class LocalFilesystemToS3Operator(BaseOperator):
    """
    Uploads a file from a local filesystem to Amazon S3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:LocalFilesystemToS3Operator`

    :param filename: Path to the local file. Path can be either absolute
            (e.g. /path/to/file.ext) or relative (e.g. ../../foo/*/*.csv). (templated)
    :param dest_key: The key of the object to copy to. (templated)

        It can be either full s3:// style url or relative path from root level.

        When it's specified as a full s3:// url, please omit `dest_bucket`.
    :param dest_bucket: Name of the S3 bucket to where the object is copied. (templated)
    :param aws_conn_id: Connection id of the S3 connection to use
    :param verify: Whether or not to verify SSL certificates for S3 connection.
        By default SSL certificates are verified.

        You can provide the following values:

        - False: do not validate SSL certificates. SSL will still be used,
                 but SSL certificates will not be
                 verified.
        - path/to/cert/bundle.pem: A filename of the CA cert bundle to uses.
                 You can specify this argument if you want to use a different
                 CA cert bundle than the one used by botocore.
    :param replace: A flag to decide whether or not to overwrite the key
            if it already exists. If replace is False and the key exists, an
            error will be raised.
    :param encrypt: If True, the file will be encrypted on the server-side
        by S3 and will be stored in an encrypted form while at rest in S3.
    :param gzip: If True, the file will be compressed locally
    :param acl_policy: String specifying the canned ACL policy for the file being
        uploaded to the S3 bucket.
    """

    template_fields: Sequence[str] = ("filename", "dest_key", "dest_bucket")

    def __init__(
        self,
        *,
        filename: str,
        dest_key: str,
        dest_bucket: str | None = None,
        aws_conn_id: str | None = "aws_default",
        verify: str | bool | None = None,
        replace: bool = False,
        encrypt: bool = False,
        gzip: bool = False,
        acl_policy: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.filename = filename
        self.dest_key = dest_key
        self.dest_bucket = dest_bucket
        self.aws_conn_id = aws_conn_id
        self.verify = verify
        self.replace = replace
        self.encrypt = encrypt
        self.gzip = gzip
        self.acl_policy = acl_policy

    def execute(self, context: Context):
        s3_hook = S3Hook(aws_conn_id=self.aws_conn_id, verify=self.verify)
        s3_bucket, s3_key = s3_hook.get_s3_bucket_key(
            self.dest_bucket, self.dest_key, "dest_bucket", "dest_key"
        )
        s3_hook.load_file(
            self.filename,
            s3_key,
            s3_bucket,
            self.replace,
            self.encrypt,
            self.gzip,
            self.acl_policy,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.batch import BatchCreateComputeEnvironmentOperator, BatchOperator
from airflow.providers.amazon.aws.operators.ecs import EcsDeregisterTaskDefinitionOperator
from airflow.providers.amazon.aws.sensors.batch import (
    BatchComputeEnvironmentSensor,
    BatchJobQueueSensor,
    BatchSensor,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import (
    ENV_ID_KEY,
    SystemTestContextBuilder,
    prune_logs,
    split_string,
)

log = logging.getLogger(__name__)

DAG_ID = "example_batch"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"
SUBNETS_KEY = "SUBNETS"
SECURITY_GROUPS_KEY = "SECURITY_GROUPS"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(ROLE_ARN_KEY)
    .add_variable(SUBNETS_KEY)
    .add_variable(SECURITY_GROUPS_KEY)
    .build()
)

JOB_OVERRIDES: dict = {}


@task
def create_job_definition(role_arn, job_definition_name):
    boto3.client("batch").register_job_definition(
        type="container",
        containerProperties={
            "command": [
                "sleep",
                "2",
            ],
            "executionRoleArn": role_arn,
            "image": "busybox",
            "resourceRequirements": [
                {"value": "1", "type": "VCPU"},
                {"value": "2048", "type": "MEMORY"},
            ],
            "networkConfiguration": {
                "assignPublicIp": "ENABLED",
            },
        },
        jobDefinitionName=job_definition_name,
        platformCapabilities=["FARGATE"],
    )


@task
def create_job_queue(job_compute_environment_name, job_queue_name):
    boto3.client("batch").create_job_queue(
        computeEnvironmentOrder=[
            {
                "computeEnvironment": job_compute_environment_name,
                "order": 1,
            },
        ],
        jobQueueName=job_queue_name,
        priority=1,
        state="ENABLED",
    )


# Only describe the job if a previous task failed, to help diagnose
@task(trigger_rule=TriggerRule.ONE_FAILED)
def describe_job(job_id):
    client = boto3.client("batch")
    response = client.describe_jobs(jobs=[job_id])
    log.info("Describing the job %s for debugging purposes", job_id)
    log.info(response["jobs"])


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_job_definition(job_definition_name):
    client = boto3.client("batch")

    response = client.describe_job_definitions(
        jobDefinitionName=job_definition_name,
        status="ACTIVE",
    )

    for job_definition in response["jobDefinitions"]:
        client.deregister_job_definition(
            jobDefinition=job_definition["jobDefinitionArn"],
        )


@task(trigger_rule=TriggerRule.ALL_DONE)
def disable_compute_environment(job_compute_environment_name):
    boto3.client("batch").update_compute_environment(
        computeEnvironment=job_compute_environment_name,
        state="DISABLED",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_compute_environment(job_compute_environment_name):
    boto3.client("batch").delete_compute_environment(
        computeEnvironment=job_compute_environment_name,
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def disable_job_queue(job_queue_name):
    boto3.client("batch").update_job_queue(
        jobQueue=job_queue_name,
        state="DISABLED",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_job_queue(job_queue_name):
    boto3.client("batch").delete_job_queue(
        jobQueue=job_queue_name,
    )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    batch_job_name: str = f"{env_id}-test-job"
    batch_job_definition_name: str = f"{env_id}-test-job-definition"
    batch_job_compute_environment_name: str = f"{env_id}-test-job-compute-environment"
    batch_job_queue_name: str = f"{env_id}-test-job-queue"

    security_groups = split_string(test_context[SECURITY_GROUPS_KEY])
    subnets = split_string(test_context[SUBNETS_KEY])

    # [START howto_operator_batch_create_compute_environment]
    create_compute_environment = BatchCreateComputeEnvironmentOperator(
        task_id="create_compute_environment",
        compute_environment_name=batch_job_compute_environment_name,
        environment_type="MANAGED",
        state="ENABLED",
        compute_resources={
            "type": "FARGATE",
            "maxvCpus": 10,
            "securityGroupIds": security_groups,
            "subnets": subnets,
        },
    )
    # [END howto_operator_batch_create_compute_environment]

    # [START howto_sensor_batch_compute_environment]
    wait_for_compute_environment_valid = BatchComputeEnvironmentSensor(
        task_id="wait_for_compute_environment_valid",
        compute_environment=batch_job_compute_environment_name,
    )
    # [END howto_sensor_batch_compute_environment]
    wait_for_compute_environment_valid.poke_interval = 1

    # [START howto_sensor_batch_job_queue]
    wait_for_job_queue_valid = BatchJobQueueSensor(
        task_id="wait_for_job_queue_valid",
        job_queue=batch_job_queue_name,
    )
    # [END howto_sensor_batch_job_queue]
    wait_for_job_queue_valid.poke_interval = 1

    # [START howto_operator_batch]
    submit_batch_job = BatchOperator(
        task_id="submit_batch_job",
        job_name=batch_job_name,
        job_queue=batch_job_queue_name,
        job_definition=batch_job_definition_name,
        container_overrides=JOB_OVERRIDES,
    )
    # [END howto_operator_batch]

    # BatchOperator waits by default, setting as False to test the Sensor below.
    submit_batch_job.wait_for_completion = False

    # [START howto_sensor_batch]
    wait_for_batch_job = BatchSensor(
        task_id="wait_for_batch_job",
        job_id=submit_batch_job.output,
    )
    # [END howto_sensor_batch]
    wait_for_batch_job.poke_interval = 10

    wait_for_compute_environment_disabled = BatchComputeEnvironmentSensor(
        task_id="wait_for_compute_environment_disabled",
        compute_environment=batch_job_compute_environment_name,
        poke_interval=1,
    )

    wait_for_job_queue_modified = BatchJobQueueSensor(
        task_id="wait_for_job_queue_modified",
        job_queue=batch_job_queue_name,
        poke_interval=1,
    )

    wait_for_job_queue_deleted = BatchJobQueueSensor(
        task_id="wait_for_job_queue_deleted",
        job_queue=batch_job_queue_name,
        treat_non_existing_as_deleted=True,
        poke_interval=10,
    )

    deregister_task_definition = EcsDeregisterTaskDefinitionOperator(
        task_id="deregister_task",
        task_definition=f"{batch_job_definition_name}:1",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            ("/aws/batch/job", env_id)
        ],
    )

    chain(
        # TEST SETUP
        test_context,
        security_groups,
        subnets,
        create_job_definition(test_context[ROLE_ARN_KEY], batch_job_definition_name),
        # TEST BODY
        create_compute_environment,
        wait_for_compute_environment_valid,
        # ``create_job_queue`` is part of test setup but need the compute-environment to be created before
        create_job_queue(batch_job_compute_environment_name, batch_job_queue_name),
        wait_for_job_queue_valid,
        submit_batch_job,
        wait_for_batch_job,
        # TEST TEARDOWN
        describe_job(submit_batch_job.output),
        disable_job_queue(batch_job_queue_name),
        wait_for_job_queue_modified,
        delete_job_queue(batch_job_queue_name),
        wait_for_job_queue_deleted,
        disable_compute_environment(batch_job_compute_environment_name),
        wait_for_compute_environment_disabled,
        delete_compute_environment(batch_job_compute_environment_name),
        delete_job_definition(batch_job_definition_name),
        deregister_task_definition,
        log_cleanup,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from boto3 import client
from moto import mock_aws

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.operators.kinesis_analytics import (
    KinesisAnalyticsV2CreateApplicationOperator,
    KinesisAnalyticsV2StartApplicationOperator,
    KinesisAnalyticsV2StopApplicationOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import BaseAwsConnection


class TestKinesisAnalyticsV2CreateApplicationOperator:
    APPLICATION_ARN = "arn:aws:kinesisanalytics:us-east-1:918313644466:application/demo"
    ROLE_ARN = "arn:aws:iam::123456789012:role/KinesisExecutionRole"

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(KinesisAnalyticsV2Hook, "conn") as _conn:
            _conn.create_application.return_value = {
                "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
            }
            yield _conn

    @pytest.fixture
    def kinesis_analytics_v2_hook(self) -> Generator[KinesisAnalyticsV2Hook, None, None]:
        with mock_aws():
            hook = KinesisAnalyticsV2Hook(aws_conn_id="aws_default")
            yield hook

    def test_init(self):
        op = KinesisAnalyticsV2CreateApplicationOperator(
            task_id="create_application_operator",
            application_name="demo",
            runtime_environment="FLINK_18_9",
            service_execution_role=self.ROLE_ARN,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        assert op.application_name == "demo"
        assert op.runtime_environment == "FLINK_18_9"
        assert op.service_execution_role == self.ROLE_ARN
        assert op.hook.client_type == "kinesisanalyticsv2"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-2"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = KinesisAnalyticsV2CreateApplicationOperator(
            task_id="create_application_operator",
            application_name="demo",
            runtime_environment="FLINK_18_9",
            service_execution_role="arn",
        )

        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_create_application(self, kinesis_analytics_mock_conn):
        self.op = KinesisAnalyticsV2CreateApplicationOperator(
            task_id="create_application_operator",
            application_name="demo",
            application_description="demo app",
            runtime_environment="FLINK_18_9",
            service_execution_role=self.ROLE_ARN,
            create_application_kwargs={
                "ApplicationConfiguration": {
                    "FlinkApplicationConfiguration": {
                        "ParallelismConfiguration": {
                            "ConfigurationType": "CUSTOM",
                            "Parallelism": 2,
                            "ParallelismPerKPU": 1,
                            "AutoScalingEnabled": True,
                        }
                    }
                }
            },
        )
        self.op.execute({})
        kinesis_analytics_mock_conn.create_application.assert_called_once_with(
            ApplicationName="demo",
            ApplicationDescription="demo app",
            RuntimeEnvironment="FLINK_18_9",
            ServiceExecutionRole=self.ROLE_ARN,
            ApplicationConfiguration={
                "FlinkApplicationConfiguration": {
                    "ParallelismConfiguration": {
                        "ConfigurationType": "CUSTOM",
                        "Parallelism": 2,
                        "ParallelismPerKPU": 1,
                        "AutoScalingEnabled": True,
                    }
                }
            },
        )

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_create_application_throw_error_when_invalid_arguments_provided(
        self, kinesis_analytics_mock_conn
    ):
        operator = KinesisAnalyticsV2CreateApplicationOperator(
            task_id="create_application_operator",
            application_name="demo",
            runtime_environment="FLINK_18_9",
            service_execution_role=self.ROLE_ARN,
            create_application_kwargs={"AppId": {"code": "s3://test/flink.jar"}},
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        operator.defer = mock.MagicMock()
        error_message = "Invalid arguments provided"

        err_response = {"Error": {"Code": "InvalidArgumentException", "Message": error_message}}

        exception = client("kinesisanalyticsv2").exceptions.ClientError(
            err_response, operation_name="CreateApplication"
        )
        returned_exception = type(exception)

        kinesis_analytics_mock_conn.exceptions.InvalidArgumentException = returned_exception
        kinesis_analytics_mock_conn.create_application.side_effect = exception

        with pytest.raises(AirflowException, match=error_message):
            operator.execute({})

    def test_template_fields(self):
        operator = KinesisAnalyticsV2CreateApplicationOperator(
            task_id="create_application_operator",
            application_name="demo",
            runtime_environment="FLINK_18_9",
            service_execution_role="arn",
        )
        validate_template_fields(operator)


class TestKinesisAnalyticsV2StartApplicationOperator:
    APPLICATION_ARN = "arn:aws:kinesisanalytics:us-east-1:123456789012:application/demo"
    ROLE_ARN = "arn:aws:iam::123456789012:role/KinesisExecutionRole"
    RUN_CONFIGURATION = {"FlinkRunConfiguration": {"AllowNonRestoredState": True}}

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(KinesisAnalyticsV2Hook, "conn") as _conn:
            _conn.start_application.return_value = {}
            _conn.describe_application.return_value = {
                "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
            }
            yield _conn

    @pytest.fixture
    def kinesis_analytics_v2_hook(self) -> Generator[KinesisAnalyticsV2Hook, None, None]:
        with mock_aws():
            hook = KinesisAnalyticsV2Hook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = KinesisAnalyticsV2StartApplicationOperator(
            task_id="start_application_operator",
            application_name="demo",
            run_configuration=self.RUN_CONFIGURATION,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        self.operator.defer = mock.MagicMock()

    def test_init(self):
        op = KinesisAnalyticsV2StartApplicationOperator(
            task_id="start_application_operator",
            application_name="demo",
            run_configuration=self.RUN_CONFIGURATION,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        assert op.application_name == "demo"
        assert op.run_configuration == self.RUN_CONFIGURATION
        assert op.hook.client_type == "kinesisanalyticsv2"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-2"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = KinesisAnalyticsV2StartApplicationOperator(
            task_id="start_application_operator",
            application_name="demo",
            run_configuration=self.RUN_CONFIGURATION,
        )

        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_start_application(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }
        kinesis_analytics_mock_conn.start_application.return_value = {}

        self.op = KinesisAnalyticsV2StartApplicationOperator(
            task_id="start_application_operator",
            application_name="demo",
            run_configuration=self.RUN_CONFIGURATION,
        )
        self.op.wait_for_completion = False
        response = self.op.execute({})

        assert response == {"ApplicationARN": self.APPLICATION_ARN}

        kinesis_analytics_mock_conn.start_application.assert_called_once_with(
            ApplicationName="demo", RunConfiguration=self.RUN_CONFIGURATION
        )

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_waiter")
    def test_start_application_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, kinesis_analytics_v2_hook
    ):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == {"ApplicationARN": self.APPLICATION_ARN}
        assert kinesis_analytics_v2_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_start_application_throw_error_when_invalid_config_provided(self, kinesis_analytics_mock_conn):
        operator = KinesisAnalyticsV2StartApplicationOperator(
            task_id="start_application_operator",
            application_name="demo",
            run_configuration={
                "ApplicationRestoreConfiguration": {
                    "ApplicationRestoreType": "SKIP_RESTORE",
                }
            },
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        operator.defer = mock.MagicMock()
        error_message = "Invalid config provided"

        err_response = {
            "Error": {"Code": "InvalidApplicationConfigurationException", "Message": error_message}
        }

        exception = client("kinesisanalyticsv2").exceptions.ClientError(
            err_response, operation_name="StartApplication"
        )
        returned_exception = type(exception)

        kinesis_analytics_mock_conn.exceptions.InvalidArgumentException = returned_exception
        kinesis_analytics_mock_conn.start_application.side_effect = exception

        with pytest.raises(AirflowException, match=error_message):
            operator.execute({})

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_execute_complete(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }

        event = {"status": "success", "application_name": "demo"}

        response = self.operator.execute_complete(context=None, event=event)

        assert response == {"ApplicationARN": self.APPLICATION_ARN}

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_execute_complete_failure(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }

        event = {"status": "error", "application_name": "demo"}

        with pytest.raises(
            AirflowException,
            match="Error while starting AWS Managed Service for Apache Flink application",
        ):
            self.operator.execute_complete(context=None, event=event)

    def test_template_fields(self):
        validate_template_fields(self.operator)


class TestKinesisAnalyticsV2StopApplicationOperator:
    APPLICATION_ARN = "arn:aws:kinesisanalytics:us-east-1:123456789012:application/demo"
    ROLE_ARN = "arn:aws:iam::123456789012:role/KinesisExecutionRole"

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(KinesisAnalyticsV2Hook, "conn") as _conn:
            _conn.stop_application.return_value = {}
            _conn.describe_application.return_value = {
                "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
            }
            yield _conn

    @pytest.fixture
    def kinesis_analytics_v2_hook(self) -> Generator[KinesisAnalyticsV2Hook, None, None]:
        with mock_aws():
            hook = KinesisAnalyticsV2Hook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = KinesisAnalyticsV2StopApplicationOperator(
            task_id="stop_application_operator",
            application_name="demo",
            force=False,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        self.operator.defer = mock.MagicMock()

    def test_init(self):
        op = KinesisAnalyticsV2StopApplicationOperator(
            task_id="stop_application_operator",
            application_name="demo",
            force=False,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        assert op.application_name == "demo"
        assert op.force is False
        assert op.hook.client_type == "kinesisanalyticsv2"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-2"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = KinesisAnalyticsV2StopApplicationOperator(
            task_id="stop_application_operator",
            application_name="demo",
            force=False,
        )

        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_stop_application(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }
        kinesis_analytics_mock_conn.stop_application.return_value = {}

        self.op = KinesisAnalyticsV2StopApplicationOperator(
            task_id="stop_application_operator", application_name="demo", force=False
        )
        self.op.wait_for_completion = False
        response = self.op.execute({})

        assert response == {"ApplicationARN": self.APPLICATION_ARN}

        kinesis_analytics_mock_conn.stop_application.assert_called_once_with(
            ApplicationName="demo", Force=False
        )

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_waiter")
    def test_stop_application_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, kinesis_analytics_v2_hook
    ):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == {"ApplicationARN": self.APPLICATION_ARN}
        assert kinesis_analytics_v2_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_stop_application_throw_error_when_invalid_config_provided(self, kinesis_analytics_mock_conn):
        operator = KinesisAnalyticsV2StopApplicationOperator(
            task_id="stop_application_operator",
            application_name="demo",
            force=False,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        operator.defer = mock.MagicMock()
        error_message = "resource not found"

        err_response = {"Error": {"Code": "ResourceNotFoundException", "Message": error_message}}

        exception = client("kinesisanalyticsv2").exceptions.ClientError(
            err_response, operation_name="StopApplication"
        )
        returned_exception = type(exception)

        kinesis_analytics_mock_conn.exceptions.ResourceNotFoundException = returned_exception
        kinesis_analytics_mock_conn.stop_application.side_effect = exception

        with pytest.raises(AirflowException, match=error_message):
            operator.execute({})

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_execute_complete(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }

        event = {"status": "success", "application_name": "demo"}

        response = self.operator.execute_complete(context=None, event=event)

        assert response == {"ApplicationARN": self.APPLICATION_ARN}

    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_execute_complete_failure(self, kinesis_analytics_mock_conn):
        kinesis_analytics_mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN}
        }
        event = {"status": "error", "application_name": "demo"}

        with pytest.raises(
            AirflowException, match="Error while stopping AWS Managed Service for Apache Flink application"
        ):
            self.operator.execute_complete(context=None, event=event)

    def test_template_fields(self):
        validate_template_fields(self.operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook
from airflow.providers.amazon.aws.sensors.glue import (
    GlueDataQualityRuleRecommendationRunSensor,
    GlueDataQualityRuleSetEvaluationRunSensor,
)

SAMPLE_RESPONSE_GET_DATA_QUALITY_EVALUATION_RUN_SUCCEEDED = {
    "RunId": "12345",
    "Status": "SUCCEEDED",
    "ResultIds": ["dqresult-123456"],
}

SAMPLE_RESPONSE_GET_DATA_QUALITY_EVALUATION_RUN_RUNNING = {
    "RunId": "12345",
    "Status": "RUNNING",
    "ResultIds": ["dqresult-123456"],
}

SAMPLE_RESPONSE_GET_DATA_QUALITY_RESULT = {
    "RunId": "12345",
    "ResultIds": ["dqresult-123456"],
    "Results": [
        {
            "ResultId": "dqresult-123456",
            "RulesetName": "rulesetOne",
            "RuleResults": [
                {
                    "Name": "Rule_1",
                    "Description": "RowCount between 150000 and 600000",
                    "EvaluatedMetrics": {"Dataset.*.RowCount": 300000.0},
                    "Result": "PASS",
                },
                {
                    "Name": "Rule_2",
                    "Description": "ColumnLength 'marketplace' between 1 and 2",
                    "EvaluationMessage": "Value: 9.0 does not meet the constraint requirement!",
                    "Result": "FAIL",
                    "EvaluatedMetrics": {
                        "Column.marketplace.MaximumLength": 9.0,
                        "Column.marketplace.MinimumLength": 2.0,
                    },
                },
            ],
        }
    ],
}

RULES = """
        Rules = [
                RowCount between 2 and 8,
                IsComplete "name"
             ]
        """

SAMPLE_RESPONSE_GET_DATA_RULE_RECOMMENDATION_RUN_SUCCEEDED = {
    "RunId": "12345",
    "Status": "SUCCEEDED",
    "DataSource": {"GlueTable": {"DatabaseName": "TestDB", "TableName": "TestTable"}},
    "RecommendedRuleset": RULES,
}

SAMPLE_RESPONSE_DATA_RULE_RECOMMENDATION_RUN_RUNNING = {"RunId": "12345", "Status": "RUNNING"}


class TestGlueDataQualityRuleSetEvaluationRunSensor:
    SENSOR = GlueDataQualityRuleSetEvaluationRunSensor

    def setup_method(self):
        self.default_args = dict(
            task_id="test_data_quality_ruleset_evaluation_run_sensor",
            evaluation_run_id="12345",
            poke_interval=5,
            max_retries=0,
        )
        self.sensor = self.SENSOR(**self.default_args, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_args)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_args,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_success_state(self, mock_conn):
        mock_conn.get_data_quality_ruleset_evaluation_run.return_value = (
            SAMPLE_RESPONSE_GET_DATA_QUALITY_EVALUATION_RUN_SUCCEEDED
        )

        assert self.sensor.poke({}) is True

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_intermediate_state(self, mock_conn):
        mock_conn.get_data_quality_ruleset_evaluation_run.return_value = (
            SAMPLE_RESPONSE_GET_DATA_QUALITY_EVALUATION_RUN_RUNNING
        )

        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.get_data_quality_ruleset_evaluation_run.return_value = {
            "RunId": "12345",
            "Status": state,
            "ResultIds": ["dqresult-123456"],
            "ErrorString": "unknown error",
        }

        sensor = self.SENSOR(**self.default_args, aws_conn_id=None)

        message = f"Error: AWS Glue data quality ruleset evaluation run RunId: 12345 Run Status: {state}: unknown error"

        with pytest.raises(AirflowException, match=message):
            sensor.poke({})

        mock_conn.get_data_quality_ruleset_evaluation_run.assert_called_once_with(RunId="12345")

    def test_sensor_defer(self):
        """Test the execute method raise TaskDeferred if running sensor in deferrable mode"""
        sensor = GlueDataQualityRuleSetEvaluationRunSensor(
            task_id="test_task",
            poke_interval=0,
            evaluation_run_id="12345",
            aws_conn_id="aws_default",
            deferrable=True,
        )

        with pytest.raises(TaskDeferred):
            sensor.execute(context=None)

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_complete_succeeds_if_status_in_succeeded_states(self, mock_conn, caplog):
        mock_conn.get_evaluation_run_results.return_value = SAMPLE_RESPONSE_GET_DATA_QUALITY_RESULT

        op = GlueDataQualityRuleSetEvaluationRunSensor(
            task_id="test_data_quality_ruleset_evaluation_run_sensor",
            evaluation_run_id="12345",
            poke_interval=0,
            aws_conn_id="aws_default",
            deferrable=True,
        )
        event = {"status": "success", "evaluation_run_id": "12345"}
        op.execute_complete(context={}, event=event)

        assert "AWS Glue data quality ruleset evaluation run completed." in caplog.messages

    def test_execute_complete_fails_if_status_in_failure_states(self):
        op = GlueDataQualityRuleSetEvaluationRunSensor(
            task_id="test_data_quality_ruleset_evaluation_run_sensor",
            evaluation_run_id="12345",
            poke_interval=0,
            aws_conn_id="aws_default",
            deferrable=True,
        )
        event = {"status": "failure"}
        with pytest.raises(AirflowException):
            op.execute_complete(context={}, event=event)


class TestGlueDataQualityRuleRecommendationRunSensor:
    SENSOR = GlueDataQualityRuleRecommendationRunSensor

    def setup_method(self):
        self.default_args = dict(
            task_id="test_data_quality_rule_recommendation_sensor",
            recommendation_run_id="12345",
            poke_interval=5,
            max_retries=0,
        )
        self.sensor = self.SENSOR(**self.default_args, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_args)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_args,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_success_state(self, mock_conn):
        mock_conn.get_data_quality_rule_recommendation_run.return_value = (
            SAMPLE_RESPONSE_GET_DATA_RULE_RECOMMENDATION_RUN_SUCCEEDED
        )
        assert self.sensor.poke({}) is True

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_intermediate_state(self, mock_conn):
        mock_conn.get_data_quality_rule_recommendation_run.return_value = (
            SAMPLE_RESPONSE_DATA_RULE_RECOMMENDATION_RUN_RUNNING
        )
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.get_data_quality_rule_recommendation_run.return_value = {
            "RunId": "12345",
            "Status": state,
            "ErrorString": "unknown error",
        }

        sensor = self.SENSOR(**self.default_args, aws_conn_id=None)

        message = (
            f"Error: AWS Glue data quality recommendation run RunId: 12345 Run Status: {state}: unknown error"
        )

        with pytest.raises(AirflowException, match=message):
            sensor.poke({})

        mock_conn.get_data_quality_rule_recommendation_run.assert_called_once_with(RunId="12345")

    def test_sensor_defer(self):
        """Test the execute method raise TaskDeferred if running sensor in deferrable mode"""
        sensor = GlueDataQualityRuleRecommendationRunSensor(
            task_id="test_task",
            poke_interval=0,
            recommendation_run_id="12345",
            aws_conn_id="aws_default",
            deferrable=True,
        )

        with pytest.raises(TaskDeferred):
            sensor.execute(context=None)

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_execute_complete_succeeds_if_status_in_succeeded_states(self, mock_conn, caplog):
        mock_conn.get_data_quality_rule_recommendation_run.return_value = (
            SAMPLE_RESPONSE_GET_DATA_RULE_RECOMMENDATION_RUN_SUCCEEDED
        )

        op = GlueDataQualityRuleRecommendationRunSensor(
            task_id="test_data_quality_rule_recommendation_run_sensor",
            recommendation_run_id="12345",
            poke_interval=0,
            aws_conn_id="aws_default",
            deferrable=True,
        )
        event = {"status": "success", "recommendation_run_id": "12345"}
        op.execute_complete(context={}, event=event)
        assert "AWS Glue data quality recommendation run completed." in caplog.messages

    def test_execute_complete_fails_if_status_in_failure_states(self):
        op = GlueDataQualityRuleRecommendationRunSensor(
            task_id="test_data_quality_rule_recommendation_run_sensor",
            recommendation_run_id="12345",
            poke_interval=0,
            aws_conn_id="aws_default",
            deferrable=True,
        )
        event = {"status": "failure"}

        with pytest.raises(AirflowException):
            op.execute_complete(context={}, event=event)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import boto3
import pytest

from airflow.providers.amazon.aws.hooks.dms import DmsHook


class TestCustomDmsWaiters:
    """Test waiters from ``amazon/aws/waiters/dms.json```."""

    @pytest.fixture(autouse=True)
    def setup_test_cases(self, monkeypatch):
        self.client = boto3.client("dms", region_name="us-east-1")
        monkeypatch.setattr(DmsHook, "conn", self.client)

    def test_service_waiters(self):
        hook_waiters = DmsHook(aws_conn_id=None).list_waiters()
        assert "replication_terminal_status" in hook_waiters
        assert "replication_config_deleted" in hook_waiters
        assert "replication_stopped" in hook_waiters
        assert "replication_complete" in hook_waiters

    @pytest.fixture
    def mock_describe_replication(self):
        with mock.patch.object(self.client, "describe_replications") as m:
            yield m

    @pytest.fixture
    def mock_describe_replication_configs(self):
        with mock.patch.object(self.client, "describe_replication_configs") as m:
            yield m

    def test_wait_for_replication_terminal_status(self, mock_describe_replication):
        mock_describe_replication.return_value = {
            "Replications": [
                {
                    "ReplicationConfigArn": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                    "Status": "failed",
                }
            ]
        }

        hook = DmsHook(aws_conn_id=None)
        waiter = hook.get_waiter("replication_terminal_status")
        waiter.wait(
            Filters=[
                {
                    "Name": "replication-instance-arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ]
        )

        mock_describe_replication.assert_called_once_with(
            Filters=[
                {
                    "Name": "replication-instance-arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ]
        )

    def test_wait_for_replication_config_delete(self, mock_describe_replication_configs):
        mock_describe_replication_configs.side_effect = [
            {"Replications": [{"ReplicationArn": "MyArn"}]},
            {"Error": {"Code": "ResourceNotFoundFault"}},
        ]

        hook = DmsHook(aws_conn_id=None)
        waiter = hook.get_waiter("replication_config_deleted")
        waiter.wait(
            Filters=[
                {
                    "Name": "replication-config-arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ],
            WaiterConfig={"Delay": 0.01, "MaxAttempts": 3},
        )

        mock_describe_replication_configs.assert_called_with(
            Filters=[
                {
                    "Name": "replication-config-arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ]
        )

    def test_wait_for_replication_stopped(self, mock_describe_replication):
        mock_describe_replication.return_value = {
            "Replications": [
                {
                    "ReplicationConfigArn": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                    "Status": "stopped",
                }
            ]
        }

        hook = DmsHook(aws_conn_id=None)
        waiter = hook.get_waiter("replication_stopped")
        waiter.wait(
            Filters=[
                {
                    "Name": "replication_config_arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ],
            WaiterConfig={"Delay": 0.01, "MaxAttempts": 3},
        )

        mock_describe_replication.assert_called_once_with(
            Filters=[
                {
                    "Name": "replication_config_arn",
                    "Values": ["XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"],
                }
            ]
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import json
from unittest import mock

import pandas as pd
from moto import mock_aws

import airflow.providers.amazon.aws.transfers.hive_to_dynamodb
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.dynamodb import DynamoDBHook

DEFAULT_DATE = datetime.datetime(2015, 1, 1)
DEFAULT_DATE_ISO = DEFAULT_DATE.isoformat()
DEFAULT_DATE_DS = DEFAULT_DATE_ISO[:10]


class TestHiveToDynamoDBOperator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        dag = DAG("test_dag_id", schedule=None, default_args=args)
        self.dag = dag
        self.sql = "SELECT 1"
        self.hook = DynamoDBHook(aws_conn_id="aws_default", region_name="us-east-1")

    @staticmethod
    def process_data(data, *args, **kwargs):
        return json.loads(data.to_json(orient="records"))

    @mock_aws
    def test_get_conn_returns_a_boto3_connection(self):
        hook = DynamoDBHook(aws_conn_id="aws_default")
        assert hook.get_conn() is not None

    @mock.patch(
        "airflow.providers.apache.hive.hooks.hive.HiveServer2Hook.get_df",
        return_value=pd.DataFrame(data=[("1", "sid")], columns=["id", "name"]),
    )
    @mock_aws
    def test_get_records_with_schema(self, mock_get_df):
        # this table needs to be created in production
        self.hook.get_conn().create_table(
            TableName="test_airflow",
            KeySchema=[
                {"AttributeName": "id", "KeyType": "HASH"},
            ],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        operator = airflow.providers.amazon.aws.transfers.hive_to_dynamodb.HiveToDynamoDBOperator(
            sql=self.sql,
            table_name="test_airflow",
            task_id="hive_to_dynamodb_check",
            table_keys=["id"],
            dag=self.dag,
        )

        operator.execute(None)

        table = self.hook.get_conn().Table("test_airflow")
        table.meta.client.get_waiter("table_exists").wait(TableName="test_airflow")
        assert table.item_count == 1

    @mock.patch(
        "airflow.providers.apache.hive.hooks.hive.HiveServer2Hook.get_df",
        return_value=pd.DataFrame(data=[("1", "sid"), ("1", "gupta")], columns=["id", "name"]),
    )
    @mock_aws
    def test_pre_process_records_with_schema(self, mock_get_df):
        # this table needs to be created in production
        self.hook.get_conn().create_table(
            TableName="test_airflow",
            KeySchema=[
                {"AttributeName": "id", "KeyType": "HASH"},
            ],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        operator = airflow.providers.amazon.aws.transfers.hive_to_dynamodb.HiveToDynamoDBOperator(
            sql=self.sql,
            table_name="test_airflow",
            task_id="hive_to_dynamodb_check",
            table_keys=["id"],
            pre_process=self.process_data,
            dag=self.dag,
        )

        operator.execute(None)

        table = self.hook.get_conn().Table("test_airflow")
        table.meta.client.get_waiter("table_exists").wait(TableName="test_airflow")
        assert table.item_count == 1



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
AWS ECS Executor configuration.

This is the configuration for calling the ECS ``run_task`` function. The AWS ECS Executor calls
Boto3's ``run_task(**kwargs)`` function with the kwargs templated by this dictionary. See the URL
below for documentation on the parameters accepted by the Boto3 run_task function.

.. seealso::
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.run_task

"""

from __future__ import annotations

import json
from json import JSONDecodeError

from airflow.configuration import conf
from airflow.providers.amazon.aws.executors.ecs.utils import (
    CONFIG_GROUP_NAME,
    ECS_LAUNCH_TYPE_EC2,
    ECS_LAUNCH_TYPE_FARGATE,
    AllEcsConfigKeys,
    RunTaskKwargsConfigKeys,
    camelize_dict_keys,
    parse_assign_public_ip,
)
from airflow.providers.amazon.aws.hooks.ecs import EcsHook
from airflow.utils.helpers import prune_dict


def _fetch_templated_kwargs() -> dict[str, str]:
    run_task_kwargs_value = conf.get(CONFIG_GROUP_NAME, AllEcsConfigKeys.RUN_TASK_KWARGS, fallback=dict())
    return json.loads(str(run_task_kwargs_value))


def _fetch_config_values() -> dict[str, str]:
    return prune_dict(
        {key: conf.get(CONFIG_GROUP_NAME, key, fallback=None) for key in RunTaskKwargsConfigKeys()}
    )


def build_task_kwargs() -> dict:
    all_config_keys = AllEcsConfigKeys()
    # This will put some kwargs at the root of the dictionary that do NOT belong there. However,
    # the code below expects them to be there and will rearrange them as necessary.
    task_kwargs = _fetch_config_values()
    task_kwargs.update(_fetch_templated_kwargs())

    has_launch_type: bool = all_config_keys.LAUNCH_TYPE in task_kwargs
    has_capacity_provider: bool = all_config_keys.CAPACITY_PROVIDER_STRATEGY in task_kwargs
    is_launch_type_ec2: bool = task_kwargs.get(all_config_keys.LAUNCH_TYPE, None) == ECS_LAUNCH_TYPE_EC2

    if has_capacity_provider and has_launch_type:
        raise ValueError(
            "capacity_provider_strategy and launch_type are mutually exclusive, you can not provide both."
        )
    if "cluster" in task_kwargs and not (has_capacity_provider or has_launch_type):
        # Default API behavior if neither is provided is to fall back on the default capacity
        # provider if it exists. Since it is not a required value, check if there is one
        # before using it, and if there is not then use the FARGATE launch_type as
        # the final fallback.
        cluster = EcsHook().conn.describe_clusters(clusters=[task_kwargs["cluster"]])["clusters"][0]
        if not cluster.get("defaultCapacityProviderStrategy"):
            task_kwargs[all_config_keys.LAUNCH_TYPE] = ECS_LAUNCH_TYPE_FARGATE

    # If you're using the EC2 launch type, you should not/can not provide the platform_version. In this
    # case we'll drop it on the floor on behalf of the user, instead of throwing an exception.
    if is_launch_type_ec2:
        task_kwargs.pop(all_config_keys.PLATFORM_VERSION, None)

    # There can only be 1 count of these containers
    task_kwargs["count"] = 1  # type: ignore
    # There could be a generic approach to the below, but likely more convoluted then just manually ensuring
    # the one nested config we need to update is present. If we need to override more options in the future we
    # should revisit this.
    if "overrides" not in task_kwargs:
        task_kwargs["overrides"] = {}  # type: ignore
    if "containerOverrides" not in task_kwargs["overrides"]:
        task_kwargs["overrides"]["containerOverrides"] = [{}]  # type: ignore
    task_kwargs["overrides"]["containerOverrides"][0]["name"] = task_kwargs.pop(  # type: ignore
        AllEcsConfigKeys.CONTAINER_NAME
    )
    # The executor will overwrite the 'command' property during execution. Must always be the first container!
    task_kwargs["overrides"]["containerOverrides"][0]["command"] = []  # type: ignore

    if any(
        [
            subnets := task_kwargs.pop(AllEcsConfigKeys.SUBNETS, None),
            security_groups := task_kwargs.pop(AllEcsConfigKeys.SECURITY_GROUPS, None),
            # Surrounding parens are for the walrus operator to function correctly along with the None check
            (assign_public_ip := task_kwargs.pop(AllEcsConfigKeys.ASSIGN_PUBLIC_IP, None)) is not None,
        ]
    ):
        network_config = prune_dict(
            {
                "awsvpcConfiguration": {
                    "subnets": str(subnets).split(",") if subnets else None,
                    "securityGroups": str(security_groups).split(",") if security_groups else None,
                    "assignPublicIp": parse_assign_public_ip(assign_public_ip, is_launch_type_ec2),
                }
            }
        )

        if "subnets" not in network_config["awsvpcConfiguration"]:
            raise ValueError("At least one subnet is required to run a task.")

        task_kwargs["networkConfiguration"] = network_config

    task_kwargs = camelize_dict_keys(task_kwargs)

    try:
        json.loads(json.dumps(task_kwargs))
    except JSONDecodeError:
        raise ValueError("AWS ECS Executor config values must be JSON serializable.")

    return task_kwargs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.lambda_function import LambdaHook
from airflow.providers.amazon.aws.sensors.lambda_function import LambdaFunctionStateSensor

FUNCTION_NAME = "function_name"


class TestLambdaFunctionStateSensor:
    def test_init(self):
        op = LambdaFunctionStateSensor(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            aws_conn_id="aws_conn_test",
            region_name="foo-bar-1",
            verify="/spam/egg.pem",
            botocore_config={"baz": "qux"},
        )

        assert op.function_name == FUNCTION_NAME

        assert op.aws_conn_id == "aws_conn_test"
        assert op.region_name == "foo-bar-1"
        assert op.verify == "/spam/egg.pem"
        assert op.botocore_config == {"baz": "qux"}

    @pytest.mark.parametrize(
        "get_function_output, expect_failure, expected",
        [
            pytest.param(
                {"Configuration": {"State": "Active"}},
                False,
                True,
                id="Active state",
            ),
            pytest.param(
                {"Configuration": {"State": "Pending"}},
                False,
                False,
                id="Pending state",
            ),
            pytest.param(
                {"Configuration": {"State": "Failed"}},
                True,
                None,
                id="Failed state",
            ),
        ],
    )
    def test_poke(self, get_function_output, expect_failure, expected):
        with mock.patch.object(LambdaHook, "conn") as mock_conn:
            mock_conn.get_function.return_value = get_function_output
            sensor = LambdaFunctionStateSensor(
                task_id="test_sensor",
                function_name=FUNCTION_NAME,
            )

            if expect_failure:
                with pytest.raises(AirflowException):
                    sensor.poke({})
            else:
                result = sensor.poke({})
                assert result == expected

            mock_conn.get_function.assert_called_once_with(
                FunctionName=FUNCTION_NAME,
            )

    def test_fail_poke(self):
        sensor = LambdaFunctionStateSensor(
            task_id="test_sensor",
            function_name=FUNCTION_NAME,
        )
        message = "Lambda function state sensor failed because the Lambda is in a failed state"
        with mock.patch("airflow.providers.amazon.aws.hooks.lambda_function.LambdaHook.conn") as conn:
            conn.get_function.return_value = {"Configuration": {"State": "Failed"}}
            with pytest.raises(AirflowException, match=message):
                sensor.poke(context={})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import boto3
import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.ses import SesHook

boto3.setup_default_session()


@mock_aws
def test_get_conn():
    hook = SesHook(aws_conn_id="aws_default")
    assert hook.get_conn() is not None


@mock_aws
@pytest.mark.parametrize(
    "to", ["to@domain.com", ["to1@domain.com", "to2@domain.com"], "to1@domain.com,to2@domain.com"]
)
@pytest.mark.parametrize(
    "cc", ["cc@domain.com", ["cc1@domain.com", "cc2@domain.com"], "cc1@domain.com,cc2@domain.com"]
)
@pytest.mark.parametrize(
    "bcc", ["bcc@domain.com", ["bcc1@domain.com", "bcc2@domain.com"], "bcc1@domain.com,bcc2@domain.com"]
)
def test_send_email(to, cc, bcc):
    # Given
    hook = SesHook()
    ses_client = hook.get_conn()
    mail_from = "test_from@domain.com"

    # Amazon only allows to send emails from verified addresses,
    # then we need to validate the from address before sending the email,
    # otherwise this test would raise a `botocore.errorfactory.MessageRejected` exception
    ses_client.verify_email_identity(EmailAddress=mail_from)

    # When
    response = hook.send_email(
        mail_from=mail_from,
        to=to,
        subject="subject",
        html_content="<html>Test</html>",
        cc=cc,
        bcc=bcc,
        reply_to="reply_to@domain.com",
        return_path="return_path@domain.com",
    )

    # Then
    assert response is not None
    assert isinstance(response, dict)
    assert "MessageId" in response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""This module contains the Amazon SageMaker Unified Studio Notebook sensor."""

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.sagemaker_unified_studio import (
    SageMakerNotebookHook,
)
from airflow.sensors.base import BaseSensorOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class SageMakerNotebookSensor(BaseSensorOperator):
    """
    Waits for a Sagemaker Workflows Notebook execution to reach any of the status below.

    'FAILED', 'STOPPED', 'COMPLETED'

    :param execution_id: The Sagemaker Workflows Notebook running execution identifier
    :param execution_name: The Sagemaker Workflows Notebook unique execution name
    """

    def __init__(self, *, execution_id: str, execution_name: str, **kwargs):
        super().__init__(**kwargs)
        self.execution_id = execution_id
        self.execution_name = execution_name
        self.success_state = ["COMPLETED"]
        self.in_progress_states = ["PENDING", "RUNNING"]

    def hook(self):
        return SageMakerNotebookHook(execution_name=self.execution_name)

    # override from base sensor
    def poke(self, context=None):
        status = self.hook().get_execution_status(execution_id=self.execution_id)

        if status in self.success_state:
            log_info_message = f"Exiting Execution {self.execution_id} State: {status}"
            self.log.info(log_info_message)
            return True
        if status in self.in_progress_states:
            return False
        error_message = f"Exiting Execution {self.execution_id} State: {status}"
        self.log.info(error_message)
        raise AirflowException(error_message)

    def execute(self, context: Context):
        # This will invoke poke method in the base sensor
        log_info_message = f"Polling Sagemaker Workflows Artifact execution: {self.execution_name} and execution id: {self.execution_id}"
        self.log.info(log_info_message)
        super().execute(context=context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow import DAG
from airflow.decorators import task, task_group
from airflow.models.baseoperator import chain
from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.operators.comprehend import (
    ComprehendCreateDocumentClassifierOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CopyObjectOperator,
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.comprehend import (
    ComprehendCreateDocumentClassifierCompletedSensor,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

ROLE_ARN_KEY = "ROLE_ARN"
BUCKET_NAME_KEY = "BUCKET_NAME"
BUCKET_KEY_DISCHARGE_KEY = "BUCKET_KEY_DISCHARGE"
BUCKET_KEY_DOCTORS_NOTES = "BUCKET_KEY_DOCTORS_NOTES"
sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(ROLE_ARN_KEY)
    .add_variable(BUCKET_NAME_KEY)
    .add_variable(BUCKET_KEY_DISCHARGE_KEY)
    .add_variable(BUCKET_KEY_DOCTORS_NOTES)
    .build()
)

DAG_ID = "example_comprehend_document_classifier"
ANNOTATION_BUCKET_KEY = "training-labels/label.csv"
TRAINING_DATA_PREFIX = "training-docs"

# Annotations file won't allow headers
# label,document name,page number

ANNOTATIONS = """DISCHARGE_SUMMARY,discharge-summary-0.pdf,1
DISCHARGE_SUMMARY,discharge-summary-1.pdf,1
DISCHARGE_SUMMARY,discharge-summary-2.pdf,1
DISCHARGE_SUMMARY,discharge-summary-3.pdf,1
DISCHARGE_SUMMARY,discharge-summary-4.pdf,1
DISCHARGE_SUMMARY,discharge-summary-5.pdf,1
DISCHARGE_SUMMARY,discharge-summary-6.pdf,1
DISCHARGE_SUMMARY,discharge-summary-7.pdf,1
DISCHARGE_SUMMARY,discharge-summary-8.pdf,1
DISCHARGE_SUMMARY,discharge-summary-9.pdf,1
DOCTOR_NOTES,doctors-notes-0.pdf,1
DOCTOR_NOTES,doctors-notes-1.pdf,1
DOCTOR_NOTES,doctors-notes-2.pdf,1
DOCTOR_NOTES,doctors-notes-3.pdf,1
DOCTOR_NOTES,doctors-notes-4.pdf,1
DOCTOR_NOTES,doctors-notes-5.pdf,1
DOCTOR_NOTES,doctors-notes-6.pdf,1
DOCTOR_NOTES,doctors-notes-7.pdf,1
DOCTOR_NOTES,doctors-notes-8.pdf,1
DOCTOR_NOTES,doctors-notes-9.pdf,1"""


@task_group
def document_classifier_workflow():
    # [START howto_operator_create_document_classifier]
    create_document_classifier = ComprehendCreateDocumentClassifierOperator(
        task_id="create_document_classifier",
        document_classifier_name=classifier_name,
        input_data_config=input_data_configurations,
        output_data_config=output_data_configurations,
        mode="MULTI_CLASS",
        data_access_role_arn=test_context[ROLE_ARN_KEY],
        language_code="en",
        document_classifier_kwargs=document_classifier_kwargs,
    )
    # [END howto_operator_create_document_classifier]
    create_document_classifier.wait_for_completion = False

    # [START howto_sensor_create_document_classifier]
    await_create_document_classifier = ComprehendCreateDocumentClassifierCompletedSensor(
        task_id="await_create_document_classifier", document_classifier_arn=create_document_classifier.output
    )
    # [END howto_sensor_create_document_classifier]

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_classifier(document_classifier_arn: str):
        ComprehendHook().conn.delete_document_classifier(DocumentClassifierArn=document_classifier_arn)

    chain(
        create_document_classifier,
        await_create_document_classifier,
        delete_classifier(create_document_classifier.output),
    )


@task
def create_kwargs_discharge():
    return [
        {
            "source_bucket_key": str(test_context[BUCKET_KEY_DISCHARGE_KEY]),
            "dest_bucket_key": f"{TRAINING_DATA_PREFIX}/discharge-summary-{counter}.pdf",
        }
        for counter in range(10)
    ]


@task
def create_kwargs_doctors_notes():
    return [
        {
            "source_bucket_key": str(test_context[BUCKET_KEY_DOCTORS_NOTES]),
            "dest_bucket_key": f"{TRAINING_DATA_PREFIX}/doctors-notes-{counter}.pdf",
        }
        for counter in range(10)
    ]


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    classifier_name = f"{env_id}-custom-document-classifier"
    bucket_name = f"{env_id}-comprehend-document-classifier"

    input_data_configurations = {
        "S3Uri": f"s3://{bucket_name}/{ANNOTATION_BUCKET_KEY}",
        "DataFormat": "COMPREHEND_CSV",
        "DocumentType": "SEMI_STRUCTURED_DOCUMENT",
        "Documents": {"S3Uri": f"s3://{bucket_name}/{TRAINING_DATA_PREFIX}/"},
        "DocumentReaderConfig": {
            "DocumentReadAction": "TEXTRACT_DETECT_DOCUMENT_TEXT",
            "DocumentReadMode": "SERVICE_DEFAULT",
        },
    }
    output_data_configurations = {"S3Uri": f"s3://{bucket_name}/output/"}
    document_classifier_kwargs = {"VersionName": "v1"}

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    discharge_kwargs = create_kwargs_discharge()
    s3_copy_discharge_task = S3CopyObjectOperator.partial(
        task_id="s3_copy_discharge_task",
        source_bucket_name=test_context[BUCKET_NAME_KEY],
        dest_bucket_name=bucket_name,
        meta_data_directive="REPLACE",
    ).expand_kwargs(discharge_kwargs)

    doctors_notes_kwargs = create_kwargs_doctors_notes()
    s3_copy_doctors_notes_task = S3CopyObjectOperator.partial(
        task_id="s3_copy_doctors_notes_task",
        source_bucket_name=test_context[BUCKET_NAME_KEY],
        dest_bucket_name=bucket_name,
        meta_data_directive="REPLACE",
    ).expand_kwargs(doctors_notes_kwargs)

    upload_annotation_file = S3CreateObjectOperator(
        task_id="upload_annotation_file",
        s3_bucket=bucket_name,
        s3_key=ANNOTATION_BUCKET_KEY,
        data=ANNOTATIONS.encode("utf-8"),
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        test_context,
        create_bucket,
        s3_copy_discharge_task,
        s3_copy_doctors_notes_task,
        upload_annotation_file,
        # TEST BODY
        document_classifier_workflow(),
        # TEST TEARDOWN
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import Any
from unittest import mock

import pendulum
import pytest

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.models import DAG, DagRun, TaskInstance
from airflow.models.variable import Variable
from airflow.providers.amazon.aws.hooks.dms import DmsHook
from airflow.providers.amazon.aws.operators.dms import (
    DmsCreateReplicationConfigOperator,
    DmsCreateTaskOperator,
    DmsDeleteReplicationConfigOperator,
    DmsDeleteTaskOperator,
    DmsDescribeReplicationConfigsOperator,
    DmsDescribeReplicationsOperator,
    DmsDescribeTasksOperator,
    DmsStartReplicationOperator,
    DmsStartTaskOperator,
    DmsStopReplicationOperator,
    DmsStopTaskOperator,
)
from airflow.providers.amazon.aws.triggers.dms import (
    DmsReplicationDeprovisionedTrigger,
    DmsReplicationTerminalStatusTrigger,
)
from airflow.utils import timezone
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS
from unit.amazon.aws.utils.test_template_fields import validate_template_fields

TASK_ARN = "test_arn"


class TestDmsCreateTaskOperator:
    TASK_DATA = {
        "replication_task_id": "task_id",
        "source_endpoint_arn": "source_endpoint",
        "target_endpoint_arn": "target_endpoint",
        "replication_instance_arn": "replication_arn",
        "table_mappings": {
            "rules": [
                {
                    "rule-type": "selection",
                    "rule-id": "1",
                    "rule-name": "1",
                    "object-locator": {
                        "schema-name": "test",
                        "table-name": "%",
                    },
                    "rule-action": "include",
                }
            ]
        },
    }

    def test_init(self):
        op = DmsCreateTaskOperator(
            task_id="create_task",
            **self.TASK_DATA,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="ca-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert op.replication_task_id == self.TASK_DATA["replication_task_id"]
        assert op.source_endpoint_arn == self.TASK_DATA["source_endpoint_arn"]
        assert op.target_endpoint_arn == self.TASK_DATA["target_endpoint_arn"]
        assert op.replication_instance_arn == self.TASK_DATA["replication_instance_arn"]
        assert op.migration_type == "full-load"
        assert op.table_mappings == self.TASK_DATA["table_mappings"]
        assert op.hook.client_type == "dms"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "ca-west-1"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = DmsCreateTaskOperator(task_id="create_task", **self.TASK_DATA)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(DmsHook, "get_task_status", side_effect=("ready",))
    @mock.patch.object(DmsHook, "create_replication_task", return_value=TASK_ARN)
    @mock.patch.object(DmsHook, "get_conn")
    def test_create_task(self, mock_conn, mock_create_replication_task, mock_get_task_status):
        dms_hook = DmsHook()

        create_task = DmsCreateTaskOperator(task_id="create_task", **self.TASK_DATA)
        create_task.execute(None)

        mock_create_replication_task.assert_called_once_with(**self.TASK_DATA, migration_type="full-load")

        assert dms_hook.get_task_status(TASK_ARN) == "ready"

    @mock.patch.object(DmsHook, "get_task_status", side_effect=("ready",))
    @mock.patch.object(DmsHook, "create_replication_task", return_value=TASK_ARN)
    @mock.patch.object(DmsHook, "get_conn")
    def test_create_task_with_migration_type(
        self, mock_conn, mock_create_replication_task, mock_get_task_status
    ):
        migration_type = "cdc"
        dms_hook = DmsHook()

        create_task = DmsCreateTaskOperator(
            task_id="create_task", migration_type=migration_type, **self.TASK_DATA
        )
        create_task.execute(None)

        mock_create_replication_task.assert_called_once_with(**self.TASK_DATA, migration_type=migration_type)

        assert dms_hook.get_task_status(TASK_ARN) == "ready"

    def test_template_fields(self):
        op = DmsCreateTaskOperator(
            task_id="create_task",
            **self.TASK_DATA,
            aws_conn_id="fake-conn-id",
            region_name="ca-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)


class TestDmsDeleteTaskOperator:
    TASK_DATA = {
        "replication_task_id": "task_id",
        "source_endpoint_arn": "source_endpoint",
        "target_endpoint_arn": "target_endpoint",
        "replication_instance_arn": "replication_arn",
        "migration_type": "full-load",
        "table_mappings": {},
    }

    def test_init(self):
        op = DmsDeleteTaskOperator(
            task_id="delete_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-east-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.replication_task_arn == TASK_ARN
        assert op.hook.client_type == "dms"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "us-east-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = DmsDeleteTaskOperator(task_id="describe_tasks", replication_task_arn=TASK_ARN)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(DmsHook, "get_task_status", side_effect=("deleting",))
    @mock.patch.object(DmsHook, "delete_replication_task")
    @mock.patch.object(DmsHook, "create_replication_task", return_value=TASK_ARN)
    @mock.patch.object(DmsHook, "get_conn")
    def test_delete_task(
        self, mock_conn, mock_create_replication_task, mock_delete_replication_task, mock_get_task_status
    ):
        dms_hook = DmsHook()
        task = dms_hook.create_replication_task(**self.TASK_DATA)

        delete_task = DmsDeleteTaskOperator(task_id="delete_task", replication_task_arn=task)
        delete_task.execute(None)

        mock_delete_replication_task.assert_called_once_with(replication_task_arn=TASK_ARN)

        assert dms_hook.get_task_status(TASK_ARN) == "deleting"

    def test_template_fields(self):
        op = DmsDeleteTaskOperator(
            task_id="delete_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-east-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)


class TestDmsDescribeTasksOperator:
    FILTER = {"Name": "replication-task-arn", "Values": [TASK_ARN]}
    MOCK_DATA = {
        "replication_task_id": "test_task",
        "source_endpoint_arn": "source-endpoint-arn",
        "target_endpoint_arn": "target-endpoint-arn",
        "replication_instance_arn": "replication-instance-arn",
        "migration_type": "full-load",
        "table_mappings": {},
    }
    MOCK_RESPONSE = [
        {
            "ReplicationTaskIdentifier": MOCK_DATA["replication_task_id"],
            "SourceEndpointArn": MOCK_DATA["source_endpoint_arn"],
            "TargetEndpointArn": MOCK_DATA["target_endpoint_arn"],
            "ReplicationInstanceArn": MOCK_DATA["replication_instance_arn"],
            "MigrationType": MOCK_DATA["migration_type"],
            "TableMappings": json.dumps(MOCK_DATA["table_mappings"]),
            "ReplicationTaskArn": TASK_ARN,
            "Status": "creating",
        }
    ]

    def setup_method(self):
        args = {
            "owner": "airflow",
            "start_date": pendulum.datetime(2018, 1, 1, tz="UTC"),
        }

        self.dag = DAG("dms_describe_tasks_operator", default_args=args, schedule="@once")

    def test_init(self):
        op = DmsDescribeTasksOperator(
            task_id="describe_tasks",
            describe_tasks_kwargs={"Filters": [self.FILTER]},
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify="/foo/bar/spam.egg",
            botocore_config={"read_timeout": 42},
        )
        assert op.describe_tasks_kwargs == {"Filters": [self.FILTER]}
        assert op.hook.client_type == "dms"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-2"
        assert op.hook._verify == "/foo/bar/spam.egg"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = DmsDescribeTasksOperator(
            task_id="describe_tasks", describe_tasks_kwargs={"Filters": [self.FILTER]}
        )
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(DmsHook, "describe_replication_tasks", return_value=(None, MOCK_RESPONSE))
    @mock.patch.object(DmsHook, "get_conn")
    def test_describe_tasks(self, mock_conn, mock_describe_replication_tasks):
        describe_tasks_kwargs = {"Filters": [self.FILTER]}
        describe_task = DmsDescribeTasksOperator(
            task_id="describe_tasks", describe_tasks_kwargs=describe_tasks_kwargs
        )
        describe_task.execute(None)

        mock_describe_replication_tasks.assert_called_once_with(**describe_tasks_kwargs)

    @pytest.mark.db_test
    @mock.patch.object(DmsHook, "describe_replication_tasks", return_value=(None, MOCK_RESPONSE))
    @mock.patch.object(DmsHook, "get_conn")
    def test_describe_tasks_return_value(self, mock_conn, mock_describe_replication_tasks, session):
        describe_task = DmsDescribeTasksOperator(
            task_id="describe_tasks", dag=self.dag, describe_tasks_kwargs={"Filters": [self.FILTER]}
        )

        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                logical_date=timezone.utcnow(),
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                execution_date=timezone.utcnow(),
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        ti = TaskInstance(task=describe_task)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        marker, response = describe_task.execute(ti.get_template_context())

        assert marker is None
        assert response == self.MOCK_RESPONSE

    def test_template_fields(self):
        op = DmsDescribeTasksOperator(
            task_id="describe_tasks",
            describe_tasks_kwargs={"Filters": [self.FILTER]},
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-2",
            verify="/foo/bar/spam.egg",
            botocore_config={"read_timeout": 42},
        )
        validate_template_fields(op)


class TestDmsStartTaskOperator:
    TASK_DATA = {
        "replication_task_id": "task_id",
        "source_endpoint_arn": "source_endpoint",
        "target_endpoint_arn": "target_endpoint",
        "replication_instance_arn": "replication_arn",
        "migration_type": "full-load",
        "table_mappings": {},
    }

    def test_init(self):
        op = DmsStartTaskOperator(
            task_id="start_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.replication_task_arn == TASK_ARN
        assert op.start_replication_task_type == "start-replication"
        assert op.hook.client_type == "dms"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "us-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = DmsStartTaskOperator(task_id="start_task", replication_task_arn=TASK_ARN)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(DmsHook, "get_task_status", side_effect=("starting",))
    @mock.patch.object(DmsHook, "start_replication_task")
    @mock.patch.object(DmsHook, "create_replication_task", return_value=TASK_ARN)
    @mock.patch.object(DmsHook, "get_conn")
    def test_start_task(
        self, mock_conn, mock_create_replication_task, mock_start_replication_task, mock_get_task_status
    ):
        dms_hook = DmsHook()
        task = dms_hook.create_replication_task(**self.TASK_DATA)

        start_task = DmsStartTaskOperator(task_id="start_task", replication_task_arn=task)
        start_task.execute(None)

        mock_start_replication_task.assert_called_once_with(
            replication_task_arn=TASK_ARN,
            start_replication_task_type="start-replication",
        )

        assert dms_hook.get_task_status(TASK_ARN) == "starting"

    def test_template_fields(self):
        op = DmsStartTaskOperator(
            task_id="start_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)


class TestDmsStopTaskOperator:
    TASK_DATA = {
        "replication_task_id": "task_id",
        "source_endpoint_arn": "source_endpoint",
        "target_endpoint_arn": "target_endpoint",
        "replication_instance_arn": "replication_arn",
        "migration_type": "full-load",
        "table_mappings": {},
    }

    def test_init(self):
        op = DmsStopTaskOperator(
            task_id="stop_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert op.replication_task_arn == TASK_ARN
        assert op.hook.client_type == "dms"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = DmsStopTaskOperator(task_id="stop_task", replication_task_arn=TASK_ARN)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(DmsHook, "get_task_status", side_effect=("stopping",))
    @mock.patch.object(DmsHook, "stop_replication_task")
    @mock.patch.object(DmsHook, "create_replication_task", return_value=TASK_ARN)
    @mock.patch.object(DmsHook, "get_conn")
    def test_stop_task(
        self, mock_conn, mock_create_replication_task, mock_stop_replication_task, mock_get_task_status
    ):
        dms_hook = DmsHook()
        task = dms_hook.create_replication_task(**self.TASK_DATA)

        stop_task = DmsStopTaskOperator(task_id="stop_task", replication_task_arn=task)
        stop_task.execute(None)

        mock_stop_replication_task.assert_called_once_with(replication_task_arn=TASK_ARN)

        assert dms_hook.get_task_status(TASK_ARN) == "stopping"

    def test_template_fields(self):
        op = DmsStopTaskOperator(
            task_id="stop_task",
            replication_task_arn=TASK_ARN,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)


class TestDmsDescribeReplicationConfigsOperator:
    filter = [{"Name": "replication-type", "Values": ["cdc"]}]

    def test_init(self):
        op = DmsDescribeReplicationConfigsOperator(task_id="test_task")
        assert op.filter is None

    @pytest.mark.db_test
    @mock.patch.object(DmsHook, "conn")
    def test_template_fields_native(self, mock_conn, session):
        logical_date = timezone.datetime(2020, 1, 1)
        Variable.set("test_filter", self.filter, session=session)

        dag = DAG(
            "test_dms",
            schedule=None,
            start_date=logical_date,
            render_template_as_native_obj=True,
        )
        op = DmsDescribeReplicationConfigsOperator(
            task_id="test_task", filter="{{ var.value.test_filter }}", dag=dag
        )

        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
                logical_date=logical_date,
            )
        else:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
                execution_date=logical_date,
            )
        ti = TaskInstance(task=op)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        context = ti.get_template_context(session)
        ti.render_templates(context)

        assert op.filter == self.filter


class TestDmsCreateReplicationConfigOperator:
    TASK_DATA = {
        "ReplicationConfigIdentifier": "test-config",
        "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
        "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
        "ComputeConfig": {
            "MaxCapacityUnits": 2,
            "MinCapacityUnits": 4,
        },
        "ReplicationType": "full-load",
        "TableMappings": json.dumps(
            {
                "TableMappings": [
                    {
                        "Type": "Selection",
                        "RuleId": 123,
                        "RuleName": "test-rule",
                        "SourceSchema": "/",
                        "SourceTable": "/",
                    }
                ]
            }
        ),
        "ReplicationSettings": "string",
        "SupplementalSettings": "string",
        "ResourceIdentifier": "string",
    }

    MOCK_REPLICATION_CONFIG_RESP: dict[str, Any] = {
        "ReplicationConfig": {
            "ReplicationConfigIdentifier": "test-config",
            "ReplicationConfigArn": "arn:aws:dms:us-east-1:123456789012:replication-config/test-config",
            "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
            "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
            "ReplicationType": "full-load",
        }
    }

    def test_init(self):
        DmsCreateReplicationConfigOperator(
            task_id="create_replication_config",
            replication_config_id=self.TASK_DATA["ReplicationConfigIdentifier"],
            source_endpoint_arn=self.TASK_DATA["SourceEndpointArn"],
            target_endpoint_arn=self.TASK_DATA["TargetEndpointArn"],
            replication_type=self.TASK_DATA["ReplicationType"],
            table_mappings=self.TASK_DATA["TableMappings"],
            compute_config=self.TASK_DATA["ComputeConfig"],
        )

    @mock.patch.object(DmsHook, "conn")
    def test_operator(self, mock_hook):
        mock_hook.create_replication_config.return_value = self.MOCK_REPLICATION_CONFIG_RESP
        op = DmsCreateReplicationConfigOperator(
            task_id="create_replication_config",
            replication_config_id=self.TASK_DATA["ReplicationConfigIdentifier"],
            source_endpoint_arn=self.TASK_DATA["SourceEndpointArn"],
            target_endpoint_arn=self.TASK_DATA["TargetEndpointArn"],
            replication_type=self.TASK_DATA["ReplicationType"],
            table_mappings=self.TASK_DATA["TableMappings"],
            compute_config=self.TASK_DATA["ComputeConfig"],
        )
        resp = op.execute(None)
        assert resp == self.MOCK_REPLICATION_CONFIG_RESP["ReplicationConfig"]["ReplicationConfigArn"]


class TestDmsDeleteReplicationConfigOperator:
    TASK_DATA = {
        "ReplicationConfigIdentifier": "test-config",
        "ReplicationConfigArn": "arn:xxxxxx",
        "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
        "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
        "ComputeConfig": {
            "MaxCapacityUnits": 2,
            "MinCapacityUnits": 4,
        },
        "ReplicationType": "full-load",
        "TableMappings": json.dumps(
            {
                "TableMappings": [
                    {
                        "Type": "Selection",
                        "RuleId": 123,
                        "RuleName": "test-rule",
                        "SourceSchema": "/",
                        "SourceTable": "/",
                    }
                ]
            }
        ),
        "ReplicationSettings": "string",
        "SupplementalSettings": "string",
        "ResourceIdentifier": "string",
    }

    def get_replication_status(self, status: str, deprovisioned: str = "deprovisioned"):
        return [
            {
                "Status": status,
                "ReplicationArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                "ReplicationIdentifier": "test-config",
                "SourceEndpointArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                "TargetEndpointArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                "ProvisionData": {"ProvisionState": deprovisioned, "ProvisionedCapacityUnits": 2},
            }
        ]

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsDeleteReplicationConfigOperator, "handle_delete_wait")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_happy_path(self, mock_waiter, mock_handle, mock_describe_replications, mock_conn):
        # testing all good statuses and no waiting
        mock_describe_replications.return_value = self.get_replication_status(
            status="stopped", deprovisioned="deprovisioned"
        )
        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=False,
        )
        op.execute({})

        mock_conn.delete_replication_config.assert_called_once()
        mock_waiter.assert_has_calls(
            [
                mock.call("replication_terminal_status"),
                mock.call().wait(
                    Filters=[{"Name": "replication-config-arn", "Values": ["arn:xxxxxx"]}],
                    WaiterConfig={"Delay": 60, "MaxAttempts": 60},
                ),
            ]
        )
        mock_handle.assert_called_once()

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    def test_defer_not_ready(self, mock_describe, mock_conn):
        mock_describe.return_value = self.get_replication_status("running")

        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, DmsReplicationTerminalStatusTrigger)

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_wait_for_completion(self, mock_waiter, mock_describe_replications, mock_conn):
        mock_describe_replications.return_value = self.get_replication_status(
            status="failed", deprovisioned="deprovisioned"
        )

        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=True,
        )
        op.execute({})

        mock_waiter.assert_has_calls(
            [
                mock.call("replication_terminal_status"),
                mock.call().wait(
                    Filters=[{"Name": "replication-config-arn", "Values": ["arn:xxxxxx"]}],
                    WaiterConfig={"Delay": 60, "MaxAttempts": 60},
                ),
            ]
        )

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_wait_for_completion_not_ready(self, mock_waiter, mock_describe_replications, mock_conn):
        mock_describe_replications.return_value = self.get_replication_status(
            status="failed", deprovisioned="xxx"
        )

        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=True,
        )
        op.execute({})

        mock_waiter.assert_has_calls(
            [
                mock.call("replication_terminal_status"),
                mock.call().wait(
                    Filters=[{"Name": "replication-config-arn", "Values": ["arn:xxxxxx"]}],
                    WaiterConfig={"Delay": 60, "MaxAttempts": 60},
                ),
            ]
        )

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsDeleteReplicationConfigOperator, "handle_delete_wait")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_not_ready_state(self, mock_waiter, mock_handle, mock_describe, mock_conn):
        mock_describe.return_value = self.get_replication_status("running")

        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=False,
        )
        op.execute({})

        mock_waiter.assert_has_calls(
            [
                mock.call("replication_terminal_status"),
                mock.call().wait(
                    Filters=[{"Name": "replication-config-arn", "Values": ["arn:xxxxxx"]}],
                    WaiterConfig={"Delay": 60, "MaxAttempts": 60},
                ),
            ]
        )
        mock_handle.assert_called_once()
        mock_conn.delete_replication_config.assert_called_once()

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsDeleteReplicationConfigOperator, "handle_delete_wait")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_not_deprovisioned(self, mock_waiter, mock_handle, mock_describe, mock_conn):
        mock_describe.return_value = self.get_replication_status("stopped", "deprovisioning")
        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=False,
        )
        op.execute({})

        mock_waiter.assert_has_calls(
            [
                mock.call("replication_terminal_status"),
                mock.call().wait(
                    Filters=[{"Name": "replication-config-arn", "Values": ["arn:xxxxxx"]}],
                    WaiterConfig={"Delay": 60, "MaxAttempts": 60},
                ),
            ]
        )
        mock_handle.assert_called_once()

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_config_not_found(self, mock_waiter, mock_describe, mock_conn):
        mock_describe.return_value = []

        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=False,
            wait_for_completion=False,
        )
        with pytest.raises(IndexError):
            op.execute({})
        mock_waiter.assert_not_called()
        mock_conn.delete_replication_config.assert_not_called()

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_defer_not_deprovisioned(self, mock_waiter, mock_describe, mock_conn):
        # not deprovisioned
        mock_describe.return_value = self.get_replication_status("stopped", "deprovisioning")
        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=True,
            wait_for_completion=False,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, DmsReplicationDeprovisionedTrigger)

        # not in terminal status
        mock_describe.return_value = self.get_replication_status("running", "deprovisioning")
        op = DmsDeleteReplicationConfigOperator(
            task_id="delete_replication_config",
            replication_config_arn=self.TASK_DATA["ReplicationConfigArn"],
            deferrable=True,
            wait_for_completion=False,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, DmsReplicationTerminalStatusTrigger)


class TestDmsDescribeReplicationsOperator:
    FILTER = [{"Name": "replication-type", "Values": ["cdc"]}]

    def get_replications(self):
        return {
            "Replications": [
                {
                    "Status": "test",
                    "ReplicationArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                    "ReplicationIdentifier": "test-config",
                    "SourceEndpointArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                    "TargetEndpointArn": "XXXXXXXXXXXXXXXXXXXXXXXXX",
                }
            ]
        }

    @mock.patch.object(DmsHook, "conn")
    def test_filter(self, mock_conn):
        mock_conn.describe_replications.return_value = self.get_replications()

        op = DmsDescribeReplicationsOperator(
            task_id="test_task",
            filter=self.FILTER,
        )

        res = op.execute({})

        mock_conn.describe_replications.assert_called_once_with(Filters=self.FILTER)
        assert isinstance(res, list)

    @mock.patch.object(DmsHook, "conn")
    def test_filter_none(self, mock_conn):
        mock_conn.describe_replications.return_value = self.get_replications()

        op = DmsDescribeReplicationsOperator(
            task_id="test_task",
        )

        res = op.execute({})

        mock_conn.describe_replications.assert_called_once_with(Filters=[])
        assert isinstance(res, list)


class TestDmsStartReplicationOperator:
    def mock_describe_replication_response(self, status: str):
        return [
            {
                "ReplicationConfigIdentifier": "string",
                "ReplicationConfigArn": "string",
                "SourceEndpointArn": "string",
                "TargetEndpointArn": "string",
                "ReplicationType": "full-load",
                "Status": status,
            }
        ]

    def mock_replication_response(self, status: str):
        return {
            "Replication": {
                "ReplicationConfigIdentifier": "xxxx",
                "ReplicationConfigArn": "xxxx",
                "Status": status,
            }
        }

    def test_arg_validation(self):
        with pytest.raises(AirflowException):
            DmsStartReplicationOperator(
                task_id="start_replication",
                replication_config_arn="XXXXXXXXXXXXXXX",
                replication_start_type="cdc",
                cdc_start_pos=1,
                cdc_start_time="2024-01-01 00:00:00",
            )
        DmsStartReplicationOperator(
            task_id="start_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            replication_start_type="cdc",
            cdc_start_pos=1,
        )

        DmsStartReplicationOperator(
            task_id="start_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            replication_start_type="cdc",
            cdc_start_time="2024-01-01 00:00:00",
        )

    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "start_replication")
    def test_already_running(self, mock_replication, mock_describe):
        mock_describe.return_value = self.mock_describe_replication_response("test")

        op = DmsStartReplicationOperator(
            task_id="start_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            replication_start_type="cdc",
            cdc_start_pos=1,
            wait_for_completion=False,
            deferrable=False,
        )

        op.execute({})
        assert mock_replication.call_count == 0

        mock_describe.return_value = self.mock_describe_replication_response("failed")
        op.execute({})
        mock_replication.return_value = self.mock_replication_response("running")
        assert mock_replication.call_count == 1

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "get_waiter")
    @mock.patch.object(DmsHook, "describe_replications")
    def test_wait_for_completion(self, mock_describe, mock_waiter, mock_conn):
        mock_describe.return_value = self.mock_describe_replication_response("stopped")
        op = DmsStartReplicationOperator(
            task_id="start_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            replication_start_type="cdc",
            cdc_start_pos=1,
            wait_for_completion=True,
            deferrable=False,
        )

        op.execute({})
        mock_waiter.assert_called_with("replication_complete")
        mock_waiter.assert_called_once()

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    def test_execute(self, mock_describe, mock_conn):
        mock_describe.return_value = self.mock_describe_replication_response("stopped")

        op = DmsStartReplicationOperator(
            task_id="start_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            replication_start_type="cdc",
            cdc_start_pos=1,
            wait_for_completion=False,
            deferrable=False,
        )
        op.execute({})
        assert mock_conn.start_replication.call_count == 1


class TestDmsStopReplicationOperator:
    def mock_describe_replication_response(self, status: str):
        return [
            {
                "ReplicationConfigIdentifier": "string",
                "ReplicationConfigArn": "string",
                "SourceEndpointArn": "string",
                "TargetEndpointArn": "string",
                "ReplicationType": "full-load",
                "Status": status,
            }
        ]

    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "conn")
    def test_already_stopped(self, mock_conn, mock_describe_replications):
        mock_describe_replications.return_value = self.mock_describe_replication_response("stopped")

        op = DmsStopReplicationOperator(
            task_id="stop_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            wait_for_completion=False,
            deferrable=False,
        )
        op.execute({})

        assert mock_conn.stop_replication.call_count == 0

    @mock.patch.object(DmsHook, "stop_replication")
    @mock.patch.object(DmsHook, "describe_replications")
    def test_execute(self, mock_describe_replications, mock_stop):
        mock_describe_replications.return_value = self.mock_describe_replication_response("started")

        op = DmsStopReplicationOperator(
            task_id="stop_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            wait_for_completion=False,
            deferrable=False,
        )
        op.execute({})
        assert mock_stop.call_count == 1

    @mock.patch.object(DmsHook, "conn")
    @mock.patch.object(DmsHook, "describe_replications")
    @mock.patch.object(DmsHook, "get_waiter")
    def test_wait_for_completion(self, mock_get_waiter, mock_describe_replications, mock_conn):
        mock_describe_replications.return_value = self.mock_describe_replication_response("started")
        op = DmsStopReplicationOperator(
            task_id="stop_replication",
            replication_config_arn="XXXXXXXXXXXXXXX",
            wait_for_completion=True,
            deferrable=False,
        )

        op.execute({})
        mock_get_waiter.assert_called_with("replication_stopped")
        mock_get_waiter.assert_called_once()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow import DAG
from airflow.models import DagRun, TaskInstance
from airflow.providers.amazon.aws.transfers.base import AwsToAwsBaseOperator
from airflow.utils import timezone
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

DEFAULT_DATE = timezone.datetime(2020, 1, 1)


class TestAwsToAwsBaseOperator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        self.dag = DAG("test_dag_id", schedule=None, default_args=args)

    @pytest.mark.db_test
    def test_render_template(self, session, clean_dags_and_dagruns):
        operator = AwsToAwsBaseOperator(
            task_id="dynamodb_to_s3_test_render",
            dag=self.dag,
            source_aws_conn_id="{{ ds }}",
            dest_aws_conn_id="{{ ds }}",
        )
        ti = TaskInstance(operator, run_id="something")
        if AIRFLOW_V_3_0_PLUS:
            ti.dag_run = DagRun(
                dag_id=self.dag.dag_id,
                run_id="something",
                logical_date=timezone.datetime(2020, 1, 1),
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            ti.dag_run = DagRun(
                dag_id=self.dag.dag_id,
                run_id="something",
                execution_date=timezone.datetime(2020, 1, 1),
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        session.add(ti)
        session.commit()
        ti.render_templates()
        assert getattr(operator, "source_aws_conn_id") == "2020-01-01"
        assert getattr(operator, "dest_aws_conn_id") == "2020-01-01"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.triggers.lambda_function import LambdaCreateFunctionCompleteTrigger


class TestLambdaCreateFunctionCompleteTrigger:
    def test_serialization(self):
        function_name = "test_function_name"
        function_arn = "test_function_arn"
        waiter_delay = 60
        waiter_max_attempts = 30
        aws_conn_id = "aws_default"

        trigger = LambdaCreateFunctionCompleteTrigger(
            function_name=function_name,
            function_arn=function_arn,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert (
            classpath
            == "airflow.providers.amazon.aws.triggers.lambda_function.LambdaCreateFunctionCompleteTrigger"
        )
        assert kwargs == {
            "function_name": "test_function_name",
            "function_arn": "test_function_arn",
            "waiter_delay": 60,
            "waiter_max_attempts": 30,
            "aws_conn_id": "aws_default",
        }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import fnmatch
import inspect
import os
import re
from collections.abc import Sequence
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable, cast

from airflow.configuration import conf
from airflow.providers.amazon.aws.utils import validate_execute_complete_event

if TYPE_CHECKING:
    from airflow.utils.context import Context

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.s3 import S3KeysUnchangedTrigger, S3KeyTrigger
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.sensors.base import poke_mode_only


class S3KeySensor(AwsBaseSensor[S3Hook]):
    """
    Waits for one or multiple keys (a file-like instance on S3) to be present in a S3 bucket.

    The path is just a key/value pointer to a resource for the given S3 path.
    Note: S3 does not support folders directly, and only provides key/value pairs.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:S3KeySensor`

    :param bucket_key: The key(s) being waited on. Supports full s3:// style url
        or relative path from root level. When it's specified as a full s3://
        url, please leave bucket_name as `None`
    :param bucket_name: Name of the S3 bucket. Only needed when ``bucket_key``
        is not provided as a full ``s3://`` url. When specified, all the keys passed to ``bucket_key``
        refers to this bucket
    :param wildcard_match: whether the bucket_key should be interpreted as a
        Unix wildcard pattern
    :param check_fn: Function that receives the list of the S3 objects with the context values,
        and returns a boolean:
        - ``True``: the criteria is met
        - ``False``: the criteria isn't met
        **Example**: Wait for any S3 object size more than 1 megabyte  ::

            def check_fn(files: List, **kwargs) -> bool:
                return any(f.get('Size', 0) > 1048576 for f in files)
    :param deferrable: Run operator in the deferrable mode
    :param use_regex: whether to use regex to check bucket
    :param metadata_keys: List of head_object attributes to gather and send to ``check_fn``.
        Acceptable values: Any top level attribute returned by s3.head_object. Specify * to return
        all available attributes.
        Default value: "Size".
        If the requested attribute is not found, the key is still included and the value is None.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    template_fields: Sequence[str] = aws_template_fields("bucket_key", "bucket_name")
    aws_hook_class = S3Hook

    def __init__(
        self,
        *,
        bucket_key: str | list[str],
        bucket_name: str | None = None,
        wildcard_match: bool = False,
        check_fn: Callable[..., bool] | None = None,
        verify: str | bool | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        use_regex: bool = False,
        metadata_keys: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.bucket_name = bucket_name
        self.bucket_key = bucket_key
        self.wildcard_match = wildcard_match
        self.check_fn = check_fn
        self.verify = verify
        self.deferrable = deferrable
        self.use_regex = use_regex
        self.metadata_keys = metadata_keys if metadata_keys else ["Size", "Key"]

    def _check_key(self, key, context: Context):
        bucket_name, key = self.hook.get_s3_bucket_key(self.bucket_name, key, "bucket_name", "bucket_key")
        self.log.info("Poking for key : s3://%s/%s", bucket_name, key)

        """
        Set variable `files` which contains a list of dict which contains attributes defined by the user
        Format: [{
            'Size': int,
            'Key': str,
        }]
        """
        if self.wildcard_match:
            prefix = re.split(r"[\[*?]", key, 1)[0]
            keys = self.hook.get_file_metadata(prefix, bucket_name)
            key_matches = [k for k in keys if fnmatch.fnmatch(k["Key"], key)]
            if not key_matches:
                return False

            # Reduce the set of metadata to requested attributes
            files = []
            for f in key_matches:
                metadata = {}
                if "*" in self.metadata_keys:
                    metadata = self.hook.head_object(f["Key"], bucket_name)
                else:
                    for key in self.metadata_keys:
                        try:
                            metadata[key] = f[key]
                        except KeyError:
                            # supplied key might be from head_object response
                            self.log.info("Key %s not found in response, performing head_object", key)
                            metadata[key] = self.hook.head_object(f["Key"], bucket_name).get(key, None)
                files.append(metadata)
        elif self.use_regex:
            keys = self.hook.get_file_metadata("", bucket_name)
            key_matches = [k for k in keys if re.match(pattern=key, string=k["Key"])]
            if not key_matches:
                return False
        else:
            obj = self.hook.head_object(key, bucket_name)
            if obj is None:
                return False
            metadata = {}
            if "*" in self.metadata_keys:
                metadata = self.hook.head_object(key, bucket_name)

            else:
                for key in self.metadata_keys:
                    # backwards compatibility with original implementation
                    if key == "Size":
                        metadata[key] = obj.get("ContentLength")
                    else:
                        metadata[key] = obj.get(key, None)
            files = [metadata]

        if self.check_fn is not None:
            # For backwards compatibility, check if the function takes a context argument
            signature = inspect.signature(self.check_fn)
            if any(param.kind == inspect.Parameter.VAR_KEYWORD for param in signature.parameters.values()):
                return self.check_fn(files, **context)
            # Otherwise, just pass the files
            return self.check_fn(files)

        return True

    def poke(self, context: Context):
        if isinstance(self.bucket_key, str):
            return self._check_key(self.bucket_key, context=context)
        return all(self._check_key(key, context=context) for key in self.bucket_key)

    def execute(self, context: Context) -> None:
        """Airflow runs this method on the worker and defers using the trigger."""
        if not self.deferrable:
            super().execute(context)
        else:
            if not self.poke(context=context):
                self._defer()

    def _defer(self) -> None:
        """Check for a keys in s3 and defers using the triggerer."""
        self.defer(
            timeout=timedelta(seconds=self.timeout),
            trigger=S3KeyTrigger(
                bucket_name=cast("str", self.bucket_name),
                bucket_key=self.bucket_key,
                wildcard_match=self.wildcard_match,
                aws_conn_id=self.aws_conn_id,
                region_name=self.region_name,
                verify=self.verify,
                botocore_config=self.botocore_config,
                poke_interval=self.poke_interval,
                should_check_fn=bool(self.check_fn),
                use_regex=self.use_regex,
            ),
            method_name="execute_complete",
        )

    def execute_complete(self, context: Context, event: dict[str, Any]) -> None:
        """
        Execute when the trigger fires - returns immediately.

        Relies on trigger to throw an exception, otherwise it assumes execution was successful.
        """
        if event["status"] == "running":
            found_keys = self.check_fn(event["files"])  # type: ignore[misc]
            if not found_keys:
                self._defer()
        elif event["status"] == "error":
            raise AirflowException(event["message"])


@poke_mode_only
class S3KeysUnchangedSensor(AwsBaseSensor[S3Hook]):
    """
    Return True if inactivity_period has passed with no increase in the number of objects matching prefix.

    Note, this sensor will not behave correctly in reschedule mode, as the state of the listed
    objects in the S3 bucket will be lost between rescheduled invocations.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:S3KeysUnchangedSensor`

    :param bucket_name: Name of the S3 bucket
    :param prefix: The prefix being waited on. Relative path from bucket root level.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param inactivity_period: The total seconds of inactivity to designate
        keys unchanged. Note, this mechanism is not real time and
        this operator may not return until a poke_interval after this period
        has passed with no additional objects sensed.
    :param min_objects: The minimum number of objects needed for keys unchanged
        sensor to be considered valid.
    :param previous_objects: The set of object ids found during the last poke.
    :param allow_delete: Should this sensor consider objects being deleted
        between pokes valid behavior. If true a warning message will be logged
        when this happens. If false an error will be raised.
    :param deferrable: Run sensor in the deferrable mode
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    template_fields: Sequence[str] = aws_template_fields("bucket_name", "prefix")
    aws_hook_class = S3Hook

    def __init__(
        self,
        *,
        bucket_name: str,
        prefix: str,
        verify: bool | str | None = None,
        inactivity_period: float = 60 * 60,
        min_objects: int = 1,
        previous_objects: set[str] | None = None,
        allow_delete: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.bucket_name = bucket_name
        self.prefix = prefix
        if inactivity_period < 0:
            raise ValueError("inactivity_period must be non-negative")
        self.inactivity_period = inactivity_period
        self.min_objects = min_objects
        self.previous_objects = previous_objects or set()
        self.inactivity_seconds = 0
        self.allow_delete = allow_delete
        self.deferrable = deferrable
        self.verify = verify
        self.last_activity_time: datetime | None = None

    def is_keys_unchanged(self, current_objects: set[str]) -> bool:
        """
        Check for new objects after the inactivity_period and update the sensor state accordingly.

        :param current_objects: set of object ids in bucket during last poke.
        """
        current_num_objects = len(current_objects)
        if current_objects > self.previous_objects:
            # When new objects arrived, reset the inactivity_seconds
            # and update previous_objects for the next poke.
            self.log.info(
                "New objects found at %s, resetting last_activity_time.",
                os.path.join(self.bucket_name, self.prefix),
            )
            self.log.debug("New objects: %s", current_objects - self.previous_objects)
            self.last_activity_time = datetime.now()
            self.inactivity_seconds = 0
            self.previous_objects = current_objects
            return False

        if self.previous_objects - current_objects:
            # During the last poke interval objects were deleted.
            if self.allow_delete:
                deleted_objects = self.previous_objects - current_objects
                self.previous_objects = current_objects
                self.last_activity_time = datetime.now()
                self.log.info(
                    "Objects were deleted during the last poke interval. Updating the "
                    "file counter and resetting last_activity_time:\n%s",
                    deleted_objects,
                )
                return False

            raise AirflowException(
                f"Illegal behavior: objects were deleted in {os.path.join(self.bucket_name, self.prefix)} between pokes."
            )

        if self.last_activity_time:
            self.inactivity_seconds = int((datetime.now() - self.last_activity_time).total_seconds())
        else:
            # Handles the first poke where last inactivity time is None.
            self.last_activity_time = datetime.now()
            self.inactivity_seconds = 0

        if self.inactivity_seconds >= self.inactivity_period:
            path = os.path.join(self.bucket_name, self.prefix)

            if current_num_objects >= self.min_objects:
                self.log.info(
                    "SUCCESS: \nSensor found %s objects at %s.\n"
                    "Waited at least %s seconds, with no new objects uploaded.",
                    current_num_objects,
                    path,
                    self.inactivity_period,
                )
                return True

            self.log.error("FAILURE: Inactivity Period passed, not enough objects found in %s", path)

            return False
        return False

    def poke(self, context: Context):
        return self.is_keys_unchanged(set(self.hook.list_keys(self.bucket_name, prefix=self.prefix)))

    def execute(self, context: Context) -> None:
        """Airflow runs this method on the worker and defers using the trigger if deferrable is True."""
        if not self.deferrable:
            super().execute(context)
        else:
            if not self.poke(context):
                self.defer(
                    timeout=timedelta(seconds=self.timeout),
                    trigger=S3KeysUnchangedTrigger(
                        bucket_name=self.bucket_name,
                        prefix=self.prefix,
                        inactivity_period=self.inactivity_period,
                        min_objects=self.min_objects,
                        previous_objects=self.previous_objects,
                        inactivity_seconds=self.inactivity_seconds,
                        allow_delete=self.allow_delete,
                        aws_conn_id=self.aws_conn_id,
                        region_name=self.region_name,
                        verify=self.verify,
                        botocore_config=self.botocore_config,
                        last_activity_time=self.last_activity_time,
                    ),
                    method_name="execute_complete",
                )

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        """
        Execute when the trigger fires - returns immediately.

        Relies on trigger to throw an exception, otherwise it assumes execution was successful.
        """
        event = validate_execute_complete_event(event)

        if event and event["status"] == "error":
            raise AirflowException(event["message"])
        return None
